/***************************************************************           

  BayCom(R)           Packet-Radio fuer IBM PC

  BayCom-Terminal
  BayCom-Node
  BayCom-Mailbox

  Routinen fuer Mailboxsystem


  ---------------------------------------------------
  Semaphoren fr den gleichzeitigen Zugriff auf Files
  ---------------------------------------------------


  Copyright (C)       Florian Radlherr
                      Taubenbergstr. 32
                      83627 Warngau

  Alle Rechte vorbehalten / All Rights Reserved

 ***************************************************************/

//19980223 OE3DZW fixed fsetopt, should not hang when file not open
//19980418 OE3DZW st sema will show option-flag
//19980925 OE3DZW dos reset-read-only only when writing to file

#include "baycom.h"

/*---------------------------------------------------------------------------*/

//
// Gesetzmaessigkeiten:
//
// Read-Zugriffe   -> beliebig viele gleichzeitig, aber nur wenn kein
//                    write-Zugriff luft
// Append-Zugriffe -> nur erlaubt, wenn sonst nur read-Zugriffe erfolgen
// Write-Zugriffe  -> nur erlaubt, wenn File sonst gar nicht im Zugriff ist
//
// Normale Zugriffe sind nach einer absehbaren Zeit beendet. D.h. es ist
// sinnvoll, den Programmablauf zu unterbrechen, und zu warten bis die
// Semaphore wieder zurckgesetzt wird. Dies ist nur bei Zugriffen sinnvoll,
// die NICHT durch eine User-Ein- oder Ausgabe blockiert werden knnen.
//
// Long-Zugriffe knnen beliebig lange dauern, das heit es lohnt nicht
// auf das Ende des Zugriffs zu warten. Das ist genau dann der Fall, wenn
// der Zugriff durch eine User-Eingabe oder durch das Vollaufen des
// Sendepuffers blockiert werden kann. Da sowas beliebig lang dauern kann,
// mu die entsprechende Stelle anders verzweigen (Aufschub des Zugriffs
// durch Pufferung (z.B. beim Lschen) oder Meldung an den Benutzer)
// option: 1=delete file on disconnect, 2=delete after close, 0 close

/*---------------------------------------------------------------------------*/

#define NO_SEMA (-1)
#undef fopen
#undef fclose

/*---------------------------------------------------------------------------*/

 typedef enum
{ s_read,
  s_longread,
  s_append,
  s_longappend,
  s_write,
  s_longwrite,
  s_maxmodes
} sema_mode;

/*---------------------------------------------------------------------------*/

typedef struct
{ char filename[FNAMELEN+1];
  char mode[s_maxmodes];
  char option; // DG4MMI: 1=delete file on disconnect, 2=delete after close
} sema_typ;

/*---------------------------------------------------------------------------*/

typedef struct
{ char name[FNAMELEN+1];
  int  tid;
} lock_typ;

/*---------------------------------------------------------------------------*/

typedef int sema_handle;

/*---------------------------------------------------------------------------*/

static sema_typ    sema[SEMA_MAX];
static sema_handle file_sema[_NFILE_];
static sema_mode   file_mode[_NFILE_];
static int         file_tid[_NFILE_];
static FILE       *file_ptr[_NFILE_];

static lock_typ    lock_sema[LOCK_MAX];

static int         hopens=0;
static int         fopens=0;
static int         hopens_max=0;
static int         fopens_max=0;

/*---------------------------------------------------------------------------*/

static sema_handle near sema_find(char *filename)
{ int i;

  for(i=0;i<SEMA_MAX;i++)
#ifdef __UNIX__ // DG4MMI
    if(strcmp(sema[i].filename,filename)==0)
#else
    if(stricmp(sema[i].filename,filename)==0)
#endif
      return i;

  return NO_SEMA;
}

/*---------------------------------------------------------------------------*/

static sema_handle near sema_new(void)
{ int i;

  for(i=0;i<SEMA_MAX;i++)
    if(sema[i].filename[0]==0)
      return i;

  return NO_SEMA;
}

/*---------------------------------------------------------------------------*/

static sema_handle near sema_set(char *filename,sema_mode mode)
{ sema_typ *s;
  int handle=sema_find(filename);

  if(handle==NO_SEMA)
  { handle=sema_new();
    if(handle==NO_SEMA)
    { trace(tr_abbruch,"s_set","semas full");
      return NO_SEMA;
    }
  }

  s=&sema[handle];

  if(s->mode[s_longwrite])
  { trace(replog,"s_set","longwrite %s",filename);
    return NO_SEMA;
  }

  switch(mode)
  {
  case s_read:
  case s_longread:
    while(s->mode[s_write] || s->mode[s_append])
    { // trace(report,"s_set","longread %s",filename);
      if(gettaskid()==NOTASK)
        return NO_SEMA;
      wdelay(528);
    }
    break;

  case s_append:
  case s_longappend:
  case s_write:
  case s_longwrite:
    if(s->mode[s_longappend])
    { trace(report,"s_set","longappend %s",filename);
      return NO_SEMA;
    }
    while(s->mode[s_write] || s->mode[s_append])
    { // trace(report,"s_set","wrd %s",filename);
      if(gettaskid()==NOTASK)
        return NO_SEMA;
      wdelay(529);
    }
    break;
  }

  if(s->filename[0]==0)
    strcpy(s->filename,filename);

  s->mode[mode]++;

  if(handle==NO_SEMA)
    trace(replog,"s_set","no sema %s",filename);

  return handle;
}

/*---------------------------------------------------------------------------*/

int sema_access(char *name)
{ int i;

  if(name==NULL)
    return 0;

  for(i=0;i<SEMA_MAX;i++)
#ifdef __UNIX__ // DG4MMI
    if(strcmp(sema[i].filename,name)==0)
#else
    if(stricmp(sema[i].filename,name)==0)
#endif
    { if(sema[i].mode[s_longwrite] || sema[i].mode[s_write])
        return 3;
      if(sema[i].mode[s_longappend] || sema[i].mode[s_append])
        return 2;
      else
        return 1;
    }

  return 0;
}

/*---------------------------------------------------------------------------*/

/*
static int sema_send(void)
{ int i;

  for(i=0;i<SEMA_MAX;i++)
    if(sema[i].mode[s_longwrite])
      return 1;

  return 0;
}
*/

/*---------------------------------------------------------------------------*/

static void near sema_clr(sema_handle handle,sema_mode mode)
{ int i;

  if((handle<SEMA_MAX) && (handle>=0) && sema[handle].filename[0])
  { if(sema[handle].mode[mode])
      sema[handle].mode[mode]--;
    else
      trace(serious,"s_clr","%s: mode %d not set",sema[handle].filename,mode);

    for(i=0;i<6;i++)
      if(sema[handle].mode[i])
        return;

    memset(&sema[handle],0,sizeof(sema_typ));
  }
  else
    trace(fatal,"s_clr","%d: no such handle",handle);
}

/*---------------------------------------------------------------------------*/

static sema_mode near s_fmode(char *mode)
{ sema_mode smode=(sema_mode)NO_SEMA;

  switch(mode[1])
  { case 'r': if(mode[2]!='+') { smode=s_read; break; }
    case 'w': smode=s_write; break;
    case 'a': smode=s_append; break;
    default:  goto error;
  }

  switch(mode[0])
  { case 'l': smode=(sema_mode)(smode+1);
    case 's': break;
    default:  goto error;
  }

  return smode;

  error: trace(tr_abbruch,"s_fmode","mode syntax %s",mode);
         return smode;
}

/*---------------------------------------------------------------------------*/

FILE *s_fopen(char *fname,char *mode)
{ sema_mode smode;
  sema_handle sh;
  FILE *f=NULL;
  handle fh;

  if(fname[0]==0)
  { trace(serious,"s_fopen","fname empty");
    return NULL;
  }  
  if (!stristr(mode,"r") || stristr(mode,"r+")) 
    resetreadonly(fname);
  smode=s_fmode(mode);
  sh=sema_set(fname,smode);
  if(sh==NO_SEMA)
  { if(b)
      b->semalock=1;
  }
  else
  { if(b)
      b->semalock=0;
    f=_fsopen(fname,mode+1,SHAREMODE);
    if(f)
    { fh=fileno(f);
      if(fh<0 || fh>_NFILE_ || file_sema[fh]!=NO_SEMA)
      { trace(fatal,"s_fopen","handle busy %s",fname);
        fclose(f);
        return NULL;
      }
      file_sema[fh]=sh;
      file_mode[fh]=smode;
      file_tid[fh]=gettaskid();
      file_ptr[fh]=f;
      fopens++;
      if(fopens>fopens_max)
        fopens_max=fopens;
    }
    else
      sema_clr(sh,smode);
  }
  //trace(report,"fopen","%d:%s",f!=NULL,fname);
  return f;
}

/*---------------------------------------------------------------------------*/

int s_set(handle fh,char *fname,char *mode)
{ sema_mode smode;
  sema_handle sh;

  smode=s_fmode(mode);
  sh=sema_set(fname,smode);
  if(sh==NO_SEMA)
  { if(b) b->semalock=1;
    trace(report,"s_set","semalock %s",fname);
  }  
  else
  { if(b) b->semalock=0;
    if(fh<0 || fh>_NFILE_ || file_sema[fh]!=NO_SEMA)
    { trace(serious,"s_set","handle busy: %d:%s",fh,fname);
      trace(serious,"s_set","old: %d:%s", file_sema[fh],
        	   	     sema[file_sema[fh]].filename);
      return EOF;
    }
    file_sema[fh]=sh;
    file_mode[fh]=smode;
    file_tid[fh]=gettaskid();
    hopens++;
    if(hopens>hopens_max)
      hopens_max=hopens;
  }
  if(b) return b->semalock;
  else  return 0;
}

/*---------------------------------------------------------------------------*/

handle s_open(char *fname,char *mode)
{ sema_mode smode;
  sema_handle  sh;
  handle fh=EOF;

  if(fname[0]==0)
  { trace(serious,"s_open","fname empty");
    return EOF;
  }
  if (!stristr(mode,"r") || stristr(mode,"r+")) 
    resetreadonly(fname);
  smode=s_fmode(mode);
  sh=sema_set(fname,smode);
  if(sh==NO_SEMA)
  { if(b) b->semalock=1;
  }
  else
  { if(b) b->semalock=0;
    switch(smode)
    { case s_read:
      case s_longread:
        fh=sopen(fname,O_RDONLY|O_BINARY,SHAREMODE);
        break;
      case s_append:
      case s_longappend:
        fh=sopen(fname,O_APPEND|O_BINARY|O_RDWR,SHAREMODE);
        if(fh!=EOF) lseek(fh,0L,SEEK_END);        // Workaround for bug
                                                  // of Borland C++ 3.1
        break;
      case s_write:
      case s_longwrite:
        if(mode[2]=='+')
        { fh=sopen(fname,O_RDWR|O_BINARY,SHAREMODE);
          if(fh==EOF)
          { if(filetime(fname))
              trace(fatal,"s_open","EOF, but file exists: %s",fname);
            fh=_creat(fname,CREATMODE);
            if(fh!=EOF)
            { _close(fh);
              fh=sopen(fname,O_RDWR|O_BINARY,SHAREMODE);
            }
          }
        }
        else
#ifdef _WIN32
         fh=sopen(fname,O_RDWR|O_BINARY|O_TRUNC|O_CREAT,SHAREMODE,CREATMODE);
#else
         fh=_creat(fname,CREATMODE);
#endif
    }
    if(fh!=EOF)
    { if(fh<0 || fh>_NFILE_ || file_sema[fh]!=NO_SEMA)
      { trace(fatal,"s_open","handle busy %s",fname);
        _close(fh);
        return EOF;
      }
      file_sema[fh]=sh;
      file_mode[fh]=smode;
      file_tid[fh]=gettaskid();
      hopens++;
      if(hopens>hopens_max)
        hopens_max=hopens;
    }
    else
      sema_clr(sh,smode);
  }
  //trace(report,"open","%d:%s",fh!=EOF,fname);
  return fh;
}

/*---------------------------------------------------------------------------*/

void s_fsetopt(FILE *f, char opt) // DG4MMI
{ if (!f) //this should never happen
    trace(fatal,"s_fsetopt","%c, no file",opt);
  s_setopt(fileno(f), opt);
}

/*---------------------------------------------------------------------------*/

void s_setopt(handle fh, char opt) // DG4MMI
{ sema_handle sh=file_sema[fh];

  if(sh==NO_SEMA)
  { trace(serious,"s_setopt","no sema set");
    return;
  }

  if(opt > sema[sh].option)
    sema[sh].option=opt;
}

/*---------------------------------------------------------------------------*/

void s_fclose(FILE *f)
{ char fname[FNAMELEN+1]="";

  if(f==NULL)
    trace(serious,"s_fclose","not open");
  else
  { handle fh=fileno(f);
    if(file_sema[fh]==NO_SEMA)
      trace(serious,"s_fclose","no sema set");
    else
    { if(fclose(f)!=0)
        trace(serious,"s_fclose","fclose %s",sema[file_sema[fh]].filename);

      if(sema[file_sema[fh]].option==2) // DG4MMI
        strcpy(fname,sema[file_sema[fh]].filename);

      sema_clr(file_sema[fh],file_mode[fh]);

      file_sema[fh]=NO_SEMA;
      file_mode[fh]=(sema_mode)NO_SEMA;
      file_tid[fh]=NOTASK;
      file_ptr[fh]=NULL;
      if(fopens)
        fopens--;
      else
        trace(serious,"s_fclose","fopens=0");

      if(fname[0])
        xunlink(fname);
    }
  }
}

/*---------------------------------------------------------------------------*/

void s_close(handle fh)
{ char fname[FNAMELEN+1]="";

  if(fh==EOF)
    trace(serious,"s_close","file not open");
  else
  { if(file_sema[fh]==NO_SEMA)
      trace(serious,"s_close","no sema set");
    else
    { if(_close(fh)!=0)
        trace(serious,"s_close","_close");

      if(sema[file_sema[fh]].option==2) // DG4MMI
        strcpy(fname,sema[file_sema[fh]].filename);

      sema_clr(file_sema[fh],file_mode[fh]);
      file_sema[fh]=NO_SEMA;
      file_mode[fh]=(sema_mode)NO_SEMA;
      file_tid[fh]=NOTASK;
      if(hopens)
        hopens--;
      else
        trace(serious,"s_close","hopens=0");

      if(fname[0])
        xunlink(fname);
    }
  }
}

/*---------------------------------------------------------------------------*/

void close_task(int tid)
{ int i;
  char fname[FNAMELEN+1]="";

  for(i=0;i<_NFILE_;i++)
    if(file_tid[i]==tid)
    { if(file_ptr[i])
      { if(sema[file_sema[i]].option==1) // DG4MMI
          strcpy(fname,sema[file_sema[i]].filename);
        s_fclose(file_ptr[i]);
      }
      else
      { if(sema[file_sema[i]].option==1) // DG4MMI
          strcpy(fname,sema[file_sema[i]].filename);
        s_close(i);
      }

      if(fname[0])
        xunlink(fname);
    }

  for(i=0;i<LOCK_MAX;i++)
  { if(lock_sema[i].tid==tid)
    { lock_sema[i].name[0]=0;
      lock_sema[i].tid=NOTASK;
    }
  }
}

/*---------------------------------------------------------------------------*/

void sema_list(void)
{ int i,j;
  putf("#   Rd LRd  Ap LAp  Wr LWr  O Name\n");
  for(i=0;i<SEMA_MAX;i++)
  { sema_typ *s=&sema[i];
    if(s->filename[0])
    { putf("%-3d",i);
      for(j=0;j<6;j++)
        putf("%2d  ",s->mode[j]);
      putf(" %d",s->option);
      putf(" %s",s->filename);
      putf("\n");
    }
  }
  putf("\nfopens=%d (max=%d), hopens=%d (max=%d).\n",fopens,fopens_max,
                                                     hopens,hopens_max);
  for(i=0;i<LOCK_MAX;i++)
  { if(lock_sema[i].name[0])
      putf("tid=%d lock=\"%s\"\n",lock_sema[i].tid,lock_sema[i].name);
  }
}

/*---------------------------------------------------------------------------*/

void openfile_list(void)
{ int i;
  static char *m[]={"Rd","LRd","Ap","LAp","Wr","LWr"};

  putf("Desc P Sema Mode Task Owner    Name\n");

  for(i=0;i<_NFILE_;i++)
  { if(file_sema[i]!=NO_SEMA)
    { putf("%3d  %c%4d%5s%5d  %-8.8s %s\n",
           i,
           file_ptr[i]?'*':' ',
           file_sema[i],
           m[file_mode[i]],
           file_tid[i],
	   gettaskname(file_tid[i]),
           sema[file_sema[i]].filename);

    }
  }
  putf("\nNFILE=%d SEMA_MAX=%d\n",_NFILE_,SEMA_MAX);
}

/*---------------------------------------------------------------------------*/

int sema_lock(char *name)
{ int i;

  for(i=0;i<LOCK_MAX;i++)
    if(stricmp(lock_sema[i].name,name)==0)
      return 0;

  for(i=0;i<LOCK_MAX;i++)
    if(lock_sema[i].name[0]==0)
    { strcpy(lock_sema[i].name,name);
      lock_sema[i].tid=gettaskid();
      return 1;
    }

  return 0;
}

/*---------------------------------------------------------------------------*/

int sema_test(char *name)
{ int i;

  for(i=0;i<LOCK_MAX;i++)
    if(stricmp(lock_sema[i].name,name)==0)
      return 1;

  return 0;
}

/*---------------------------------------------------------------------------*/

void sema_unlock(char *name)
{ int tid=gettaskid();
  int i;

  for(i=0;i<LOCK_MAX;i++)
    if(stricmp(lock_sema[i].name,name)==0 && lock_sema[i].tid==tid)
    { lock_sema[i].name[0]=0;
      lock_sema[i].tid=NOTASK;
    }
}

/*---------------------------------------------------------------------------*/

void sema_init(void)
{ int i;

  memset(sema,0,sizeof(sema_typ)*SEMA_MAX);

  for(i=0;i<_NFILE_;i++)
  { file_sema[i]=NO_SEMA;
    file_mode[i]=(sema_mode)NO_SEMA;
    file_tid[i]=NOTASK;
    file_ptr[i]=NULL;
  }

  for(i=0;i<LOCK_MAX;i++)
  { lock_sema[i].name[0]=0;
    lock_sema[i].tid=NOTASK;
  }
}

/*---------------------------------------------------------------------------*/
